"""
Test basic flow
"""
import datetime
from collections import namedtuple
import asyncio

import pytest

from aio_sf_streaming import SimpleSalesforceStreaming
from ..utils.async_itertools import async_enumerate
from ..utils.async_tools import wait_until_all_completed


# ==================== Test client ====================

class Replay(namedtuple('Replay', 'channel replay_id creation_time')):
    """
    Record a Replay message
    """


class MyTestClient(SimpleSalesforceStreaming):
    def __init__(self, **kwargs):
        self.stored_replay = set()
        super().__init__(**kwargs)

    async def store_replay_id(self, channel, replay_id, creation_time):
        self.stored_replay.add(Replay(channel, replay_id, creation_time))

    async def get_last_replay_id(self, channel):
        elements = {r for r in self.stored_replay if r.channel == channel}
        if not elements:
            return None
        return max(elements, key=lambda r: r.creation_time)


async def send_events(count, sleep, channel, server):
    """
    Allow to send fake event to the server
    """
    for i in range(count):
        await asyncio.sleep(sleep)
        server.push_result(channel, {"foo": i})


@pytest.mark.asyncio
async def test_basic_flow(event_loop, fake_sf_session):
    # Launch our streaming client for few events

    received_messages = []
    async with MyTestClient(username="my-username",
                            password="my-password",
                            client_id="my-client_id",
                            client_secret="my-client_secret",
                            loop=event_loop,
                            login_connector=fake_sf_session['login_connector'],
                            connector=fake_sf_session['connector']) as sfs:
        await sfs.subscribe('/topic/Foo')
        await sfs.subscribe('/topic/Bar')
        # Push messages from server
        event_loop.create_task(send_events(2, 1, '/topic/Foo', fake_sf_session['sf_server']))
        event_loop.create_task(send_events(2, 1.5, '/topic/Bar', fake_sf_session['sf_server']))

        async for i, message in async_enumerate(sfs.events()):
            received_messages.append(message)
            if i == 3:
                await sfs.unsubscribe('/topic/Foo')
                await sfs.unsubscribe('/topic/Bar')
                await sfs.ask_stop()

    await wait_until_all_completed()

    # check results
    assert received_messages == [{
        'data': {
            'sobject': {
                'foo': 0
            },
            'event': {
                'replayId': 1,
                'createdDate': '2018-03-15T13:42:01'
            }
        },
        'channel': '/topic/Foo'
    }, {
        'data': {
            'sobject': {
                'foo': 0
            },
            'event': {
                'replayId': 1,
                'createdDate': '2018-03-15T13:42:02'
            }
        },
        'channel': '/topic/Bar'
    }, {
        'data': {
            'sobject': {
                'foo': 1
            },
            'event': {
                'replayId': 2,
                'createdDate': '2018-03-15T13:42:03'
            }
        },
        'channel': '/topic/Foo'
    }, {
        'data': {
            'sobject': {
                'foo': 1
            },
            'event': {
                'replayId': 2,
                'createdDate': '2018-03-15T13:42:04'
            }
        },
        'channel': '/topic/Bar'
    }]

    assert fake_sf_session['sf_server'].received_messages == [
        {
            'channel': '/meta/handshake',
            'supportedConnectionTypes': ['long-polling'],
            'version': '1.0',
            'minimumVersion': '1.0',
            'ext': {
                'replay': True
            },
            'id': '1'
        }, {
            'channel': '/meta/subscribe',
            'subscription': '/topic/Foo',
            'ext': {
                'replay': {
                    '/topic/Foo': -1
                }
            },
            'id': '2',
            'clientId': '4'
        }, {
            'channel': '/meta/subscribe',
            'subscription': '/topic/Bar',
            'ext': {
                'replay': {
                    '/topic/Bar': -1
                }
            },
            'id': '3',
            'clientId': '4'
        }, {
            'channel': '/meta/connect',
            'connectionType': 'long-polling',
            'id': '4',
            'clientId': '4'
        }, {
            'channel': '/meta/connect',
            'connectionType': 'long-polling',
            'id': '5',
            'clientId': '4'
        }, {
            'channel': '/meta/connect',
            'connectionType': 'long-polling',
            'id': '6',
            'clientId': '4'
        }, {
            'channel': '/meta/connect',
            'connectionType': 'long-polling',
            'id': '7',
            'clientId': '4'
        }, {
            'channel': '/meta/connect',
            'connectionType': 'long-polling',
            'id': '8',
            'clientId': '4'
        }, {
            'channel': '/meta/unsubscribe',
            'subscription': '/topic/Foo',
            'id': '9',
            'clientId': '4'
        }, {
            'channel': '/meta/unsubscribe',
            'subscription': '/topic/Bar',
            'id': '10',
            'clientId': '4'
        }, {
            'channel': '/meta/disconnect',
            'id': '11',
            'clientId': '4'
        }]

    assert sfs.stored_replay == {
        Replay(
            channel='/topic/Foo',
            replay_id=1,
            creation_time='2018-03-15T13:42:01'),
        Replay(
            channel='/topic/Bar',
            replay_id=2,
            creation_time='2018-03-15T13:42:04'),
        Replay(
            channel='/topic/Bar',
            replay_id=1,
            creation_time='2018-03-15T13:42:02'),
        Replay(
            channel='/topic/Foo',
            replay_id=2,
            creation_time='2018-03-15T13:42:03')
    }



"""
Unit tests for connectors flow
"""

import asyncio
import datetime as dt
from math import isclose
from unittest.mock import call

from asynctest import patch, CoroutineMock
import pytest

from aio_sf_streaming import (
    BaseSalesforceStreaming,
    TimeoutAdviceMixin,
    ReplayMixin,
    ReplayType,
    AutoVersionMixin,
    AutoReconnectMixin,
    ReSubscribeMixin,
)
from ..utils.async_itertools import async_enumerate
from ..utils.async_tools import wait_until_all_completed


class TimeoutTestClass(TimeoutAdviceMixin, BaseSalesforceStreaming):
    """
    A fake sf streaming derivated class that always return a fake token
    """

    TEST_ACCESS_TOKEN = "42"
    TEST_INSTANCE_URL = "https://my-instance.com"

    async def fetch_token(self):
        return self.TEST_ACCESS_TOKEN, self.TEST_INSTANCE_URL


@patch("aio_sf_streaming.BaseSalesforceStreaming.send")
@pytest.mark.asyncio
async def test_timeout_advice(mock_send):
    """
    Test timeout advice mixin
    """

    messages = [
        {"channel": "/meta/connect", "advice": {"timeout": 42000}},
        {"channel": "/topic/Foo0"},
        {"channel": "/topic/Foo1"},
        {"channel": "/topic/Foo2"},
        {"channel": "/topic/Foo3"},
        {"channel": "/topic/Foo4"},
        {"channel": "/meta/connect"},
        {"channel": "/topic/Foo5"},
        {"channel": "/meta/connect"},
    ]
    mock_send.side_effect = [
        # First message : a simple success
        [messages[0]],
        # Second message : multiples messages
        messages[1:4],
        # Timeout
        asyncio.TimeoutError(),
        # Other messages
        messages[4:],
    ]
    client = TimeoutTestClass()
    old_timeout = client.timeout
    received_messages = []
    # We iterate message and ask stop after received 7 messages
    async for i, m in async_enumerate(client.messages()):
        received_messages.append(m)
        if i == 6:
            await client.ask_stop()

    # We should have only 7 messages
    assert received_messages == messages[:7]
    assert mock_send.call_count == 4

    # Timeout advice should be used
    new_timeout = client.timeout
    assert new_timeout is not old_timeout
    assert isclose(new_timeout, 42.)


class ReplayTestClass(ReplayMixin, BaseSalesforceStreaming):
    """
    A fake sf streaming derivated class that always return a fake token
    """

    TEST_ACCESS_TOKEN = "42"
    TEST_INSTANCE_URL = "https://my-instance.com"

    async def fetch_token(self):
        return self.TEST_ACCESS_TOKEN, self.TEST_INSTANCE_URL


@pytest.mark.asyncio
async def test_replay_handshake_payload():
    """
    Test replay handshake_payload
    """
    client = ReplayTestClass()
    # the handshake payload is constant
    payload = await client.get_handshake_payload()

    assert "ext" in payload
    assert payload["ext"].get("replay") is True


@patch.object(ReplayTestClass, "get_last_replay_id")
@pytest.mark.asyncio
async def test_replay_subscribe_payload(mock_last_id):
    """
    Test replay subscribe_payload
    """
    client = ReplayTestClass()

    # By default, new events
    mock_last_id.return_value = None
    mock_last_id.reset_mock()
    result = await client.get_subscribe_payload("/topic/Foo")
    assert mock_last_id.call_count == 1
    assert result["ext"]["replay"]["/topic/Foo"] == -1

    # ReplayType.NEW_EVENTS => -1
    mock_last_id.return_value = ReplayType.NEW_EVENTS
    mock_last_id.reset_mock()
    result = await client.get_subscribe_payload("/topic/Foo")
    assert mock_last_id.call_count == 1
    assert result["ext"]["replay"]["/topic/Foo"] == -1

    # ReplayType.ALL_EVENTS => -2
    mock_last_id.return_value = ReplayType.ALL_EVENTS
    mock_last_id.reset_mock()
    result = await client.get_subscribe_payload("/topic/Foo")
    assert mock_last_id.call_count == 1
    assert result["ext"]["replay"]["/topic/Foo"] == -2

    # Specific value : used directly
    mock_last_id.return_value = 42
    mock_last_id.reset_mock()
    result = await client.get_subscribe_payload("/topic/Foo")
    assert mock_last_id.call_count == 1
    assert result["ext"]["replay"]["/topic/Foo"] == 42


@patch.object(ReplayTestClass, "store_replay_id")
@patch("aio_sf_streaming.BaseSalesforceStreaming.send")
@pytest.mark.asyncio
async def test_replay_message(mock_send, mock_store_replay):
    """
    Test replay message: should send replay id to a the callback
    """

    messages = [
        {
            "channel": "/topic/Foo0",
            "data": {"event": {"replayId": 1, "createdDate": "2018-03-14T11:58:42.1234Z"}},
        },
        {
            "channel": "/topic/Foo1",
            "data": {"event": {"replayId": 1, "createdDate": "2018-03-14T11:58:43.1234Z"}},
        },
        {
            "channel": "/topic/Foo0",
            "data": {"event": {"replayId": 2, "createdDate": "2018-03-14T11:58:44.1234Z"}},
        },
        {
            "channel": "/topic/Foo1",
            "data": {"event": {"replayId": 2, "createdDate": "2018-03-14T11:58:45.1234Z"}},
        },
        {
            "channel": "/topic/Foo2",
            "data": {"event": {"replayId": 1, "createdDate": "2018-03-14T11:58:46.1234Z"}},
        },
        {
            "channel": "/topic/Foo0",
            "data": {"event": {"replayId": 3, "createdDate": "2018-03-14T11:58:47.1234Z"}},
        },
        {
            "channel": "/topic/Foo0",
            "data": {"event": {"replayId": 4, "createdDate": "2018-03-14T11:58:48.1234Z"}},
        },
        {
            "channel": "/topic/Foo1",
            "data": {"event": {"replayId": 3, "createdDate": "2018-03-14T11:58:49.1234Z"}},
        },
        {
            "channel": "/topic/Foo0",
            "data": {"event": {"replayId": 5, "createdDate": "2018-03-14T11:58:50.1234Z"}},
        },
        {
            "channel": "/topic/Foo1",
            "data": {"event": {"replayId": 4, "createdDate": "2018-03-14T11:58:51.1234Z"}},
        },
        {
            "channel": "/topic/Foo2",
            "data": {"event": {"replayId": 2, "createdDate": "2018-03-14T11:58:52.1234Z"}},
        },
        {
            "channel": "/topic/Foo0",
            "data": {"event": {"replayId": 6, "createdDate": "2018-03-14T11:58:53.1234Z"}},
        },
        {"channel": "/meta/connect"},
        {"channel": "/meta/connect"},
    ]
    mock_send.side_effect = [
        # First message : a simple success
        [messages[0]],
        # Second message : multiples messages
        messages[1:4],
        # Other messages
        messages[4:-2],
        [messages[-2]],
        [messages[-1]],
    ]
    client = ReplayTestClass()
    async for i, m in async_enumerate(client.messages()):
        assert m == messages[i]
        if i == len(messages) - 2:
            await client.ask_stop()
    await wait_until_all_completed()
    assert mock_store_replay.call_count == 12
    assert mock_store_replay.mock_calls == [
        call("/topic/Foo0", 1, "2018-03-14T11:58:42.1234Z"),
        call("/topic/Foo1", 1, "2018-03-14T11:58:43.1234Z"),
        call("/topic/Foo0", 2, "2018-03-14T11:58:44.1234Z"),
        call("/topic/Foo1", 2, "2018-03-14T11:58:45.1234Z"),
        call("/topic/Foo2", 1, "2018-03-14T11:58:46.1234Z"),
        call("/topic/Foo0", 3, "2018-03-14T11:58:47.1234Z"),
        call("/topic/Foo0", 4, "2018-03-14T11:58:48.1234Z"),
        call("/topic/Foo1", 3, "2018-03-14T11:58:49.1234Z"),
        call("/topic/Foo0", 5, "2018-03-14T11:58:50.1234Z"),
        call("/topic/Foo1", 4, "2018-03-14T11:58:51.1234Z"),
        call("/topic/Foo2", 2, "2018-03-14T11:58:52.1234Z"),
        call("/topic/Foo0", 6, "2018-03-14T11:58:53.1234Z"),
    ]


class AutoVersionTestClass(AutoVersionMixin, BaseSalesforceStreaming):
    """
    A fake sf streaming derivated class that always return a fake token
    """

    TEST_ACCESS_TOKEN = "42"
    TEST_INSTANCE_URL = "https://my-instance.com"

    async def fetch_token(self):
        return self.TEST_ACCESS_TOKEN, self.TEST_INSTANCE_URL


@patch("aio_sf_streaming.BaseSalesforceStreaming.handshake")
@patch("aio_sf_streaming.BaseSalesforceStreaming.get")
@pytest.mark.asyncio
async def test_auto_version(mock_get, mock_handshake):
    """
    Test auto version test
    """
    mock_get.return_value = []
    client = AutoVersionTestClass(version="1.0")

    # get should be called but version not updated.
    await client.handshake()
    assert mock_get.call_count == 1
    assert mock_get.call_args == call("/services/data/")
    assert client.version == "1.0"
    assert mock_handshake.call_count == 1

    mock_get.return_value = [{"version": "42.0"}]
    mock_get.reset_mock()
    mock_handshake.reset_mock()

    await client.handshake()
    assert mock_get.call_count == 1
    assert mock_get.call_args == call("/services/data/")
    assert client.version == "42.0"
    assert mock_handshake.call_count == 1


class AutoReconnectTestClass(AutoReconnectMixin, BaseSalesforceStreaming):
    """
    A fake sf streaming derivated class that always return a fake token
    """

    TEST_ACCESS_TOKEN = "42"
    TEST_INSTANCE_URL = "https://my-instance.com"

    async def fetch_token(self):
        return self.TEST_ACCESS_TOKEN, self.TEST_INSTANCE_URL


@patch("aio_sf_streaming.BaseSalesforceStreaming.start")
@patch("aio_sf_streaming.BaseSalesforceStreaming.unsubscribe")
@patch("aio_sf_streaming.BaseSalesforceStreaming.stop")
@patch("aio_sf_streaming.BaseSalesforceStreaming.handshake")
@patch("aio_sf_streaming.BaseSalesforceStreaming.subscribe")
@patch("aio_sf_streaming.BaseSalesforceStreaming.send")
@pytest.mark.asyncio
async def test_auto_reconnect(mock_send, mock_subscribe, mock_handshake, *_):
    """
    Test auto reconnect
    """
    client = AutoReconnectTestClass()
    await client.start()
    await client.subscribe("/topic/Foo")
    await client.subscribe("/topic/Bar")
    await client.subscribe("/topic/Baz")

    messages = [
        {"channel": "/topic/Foo"},
        {"channel": "/topic/Foo"},
        {"channel": "/topic/Bar"},
        {"channel": "/topic/Foo"},
        {"channel": "/topic/Baz"},
        {"channel": "/meta/connect", "error": "403::Unknown client"},
        {"channel": "/topic/Baz"},
        {"channel": "/topic/Foo"},
        {"channel": "/topic/Foo"},
        {"channel": "/topic/Bar"},
        {"channel": "/topic/Foo"},
        {"channel": "/meta/connect", "error": "403::Unknown client"},
        {"channel": "/topic/Foo"},
        {"channel": "/topic/Foo"},
        {"channel": "/topic/Bar"},
        {"channel": "/topic/Foo"},
        {"channel": "/meta/connect"},
        {"channel": "/meta/connect"},
    ]
    mock_send.side_effect = [[message] for message in messages]
    j = 0
    async for i, m in async_enumerate(client.messages()):
        if i == 0:
            # First step, only subscribe from test, no specific handhshake
            assert mock_subscribe.call_count == 3
            mock_subscribe.reset_mock()
            assert mock_handshake.call_count == 0
        elif i == 4:
            # nothing new before disconnect
            assert mock_subscribe.call_count == 0
            assert mock_handshake.call_count == 0
        elif i == 5:
            # After a disconnect, handshake and new subscribe has been call
            await wait_until_all_completed()
            assert mock_subscribe.call_count == 3
            assert call("/topic/Foo") in mock_subscribe.mock_calls
            assert call("/topic/Bar") in mock_subscribe.mock_calls
            assert call("/topic/Baz") in mock_subscribe.mock_calls
            mock_subscribe.reset_mock()
            assert mock_handshake.call_count == 1
            mock_handshake.reset_mock()
            j += 1
        elif i == 7:
            # Here, we unsubscribe
            await client.unsubscribe("/topic/Baz")
        elif i == 9:
            # nothing new before disconnect
            assert mock_subscribe.call_count == 0
            assert mock_handshake.call_count == 0
        elif i == 10:
            # After a disconnect, handshake and new subscribe has been call
            await wait_until_all_completed()
            assert mock_subscribe.call_count == 2
            assert call("/topic/Foo") in mock_subscribe.mock_calls
            assert call("/topic/Bar") in mock_subscribe.mock_calls
            mock_subscribe.reset_mock()
            assert mock_handshake.call_count == 1
            mock_handshake.reset_mock()
            j += 1

        assert m == messages[j]
        j += 1
        if i == len(messages) - 4:
            await client.ask_stop()
    await client.stop()


class ReSubscribeTestClass(ReSubscribeMixin, BaseSalesforceStreaming):
    """
    A fake sf streaming derivated class that always return a fake token
    """

    TEST_ACCESS_TOKEN = "42"
    TEST_INSTANCE_URL = "https://my-instance.com"

    async def fetch_token(self):
        return self.TEST_ACCESS_TOKEN, self.TEST_INSTANCE_URL


@patch("aio_sf_streaming.BaseSalesforceStreaming.subscribe")
@pytest.mark.asyncio
async def test_resubscribe(mock_subscribe):
    """
    Test re-subscribe
    """
    mock_subscribe.side_effect = [
        [{"successful": False, "ext": {"sfdc": {"failureReason": "SERVER_UNAVAILABLE"}}}],
        [{"successful": False, "ext": {"sfdc": {"failureReason": "SERVER_UNAVAILABLE"}}}],
        [{"successful": True}],
    ]
    client = ReSubscribeTestClass()
    response = await client.subscribe("/topic/Foo")

    assert response[0]["successful"]
    assert mock_subscribe.call_count == 3


class ReSubscribeCallBackTestClass(ReSubscribeMixin, BaseSalesforceStreaming):
    """
    A fake sf streaming derivated class that always return a fake token
    """

    TEST_ACCESS_TOKEN = "42"
    TEST_INSTANCE_URL = "https://my-instance.com"

    async def fetch_token(self):
        return self.TEST_ACCESS_TOKEN, self.TEST_INSTANCE_URL

    async def should_retry_on_exception(self, channel, exception):
        return isinstance(exception, ValueError)

    async def should_retry_on_error_response(self, channel, response):
        return False


@patch("aio_sf_streaming.BaseSalesforceStreaming.subscribe")
@pytest.mark.asyncio
async def test_resubscribe_callback(mock_subscribe):
    """
    Test re-subscribe
    """
    mock_subscribe.side_effect = [
        ValueError(),
        ValueError(),
        [{"successful": False, "ext": {"sfdc": {"failureReason": "SERVER_UNAVAILABLE"}}}],
    ]
    client = ReSubscribeCallBackTestClass()

    response = await client.subscribe("/topic/Foo")

    assert not response[0]["successful"]
    assert mock_subscribe.call_count == 3



"""
Some pytest fixture helper
"""
import aiohttp
import pytest

from .utils.fake_server import FakeLoginSfServer, FakeResolver, FakeSfServer


@pytest.fixture()
async def fake_login_server(event_loop):
    """
    Fixture: create a fake login server
    """
    fake_sf = FakeLoginSfServer(loop=event_loop)
    await fake_sf.start()
    yield fake_sf
    await fake_sf.stop()


@pytest.fixture()
async def fake_sf_server(event_loop):
    """
    Fixture: create a fake sf server
    """
    fake_sf = FakeSfServer(loop=event_loop)
    await fake_sf.start()
    yield fake_sf
    await fake_sf.stop()


@pytest.fixture()
async def fake_sf_session(event_loop, fake_sf_server, fake_login_server):
    """
    Fixture: create a fake sf session
    """
    # Merge all fake server hosts
    info = {**fake_sf_server.host_mapping, **fake_login_server.host_mapping}
    resolver = FakeResolver(info, loop=event_loop)
    # We need one connector by connection
    connector = aiohttp.TCPConnector(
        loop=event_loop, resolver=resolver, verify_ssl=False
    )
    login_connector = aiohttp.TCPConnector(
        loop=event_loop, resolver=resolver, verify_ssl=False
    )
    # Return all data
    yield {
        "login_connector": login_connector,
        "connector": connector,
        "login_server": fake_login_server,
        "sf_server": fake_sf_server,
    }



"""
Various async utilities functions
"""
import asyncio


async def wait_until_all_completed():
    """
    Small utility that check pending tasks and return when theire is only one
    (= the active task)
    """
    pending = asyncio.Task.all_tasks()
    while "not all tasks completed":
        _, pending = await asyncio.wait(pending,
                                        return_when=asyncio.FIRST_COMPLETED)
        if len(pending) == 1:
            return



"""
Fake sf server

We create a fake resolver and servers allowing to simulate SF logic for our
tests.
Heavily inspired from:
 - <https://github.com/aio-libs/aiohttp/blob/master/examples/fake_server.py>
 - <https://gist.github.com/ambivalentno/e311ea008d05938ac5dd3048ce76e3d1>
"""
import datetime
import asyncio
from collections import deque, defaultdict
import pathlib
import ssl
import socket
from aiohttp import web
from aiohttp.resolver import DefaultResolver
from aiohttp.test_utils import unused_port


class FakeResolver:
    """
    Fake resolver will redirect provider host to a specific local port
    """
    _LOCAL_HOST = {0: '127.0.0.1',
                   socket.AF_INET: '127.0.0.1',
                   socket.AF_INET6: '::1'}

    def __init__(self, fakes, *, loop):
        """fakes -- dns -> port dict"""
        self._fakes = fakes
        self._resolver = DefaultResolver(loop=loop)

    async def resolve(self, host, port=0, family=socket.AF_INET):
        """
        Resolve an host: check if a fake server is registred and redirect to it.
        If the host is unknown, raise an error
        """
        fake_port = self._fakes.get(host)
        if fake_port is not None:
            return [{'hostname': host,
                     'host': self._LOCAL_HOST[family], 'port': fake_port,
                     'family': family, 'proto': 0,
                     'flags': socket.AI_NUMERICHOST}]
        raise OSError("DNS lookup failed")


class BaseFakeServer:
    """
    Base class for fake SF server
    """
    def __init__(self, *, loop):
        self.loop = loop
        self.app = web.Application(loop=loop)
        self.handler = None
        self.server = None
        self.port = None
        self.host = None
        # Use fake certificate for https
        here = pathlib.Path(__file__)
        ssl_cert = here.parent / 'server.crt'
        ssl_key = here.parent / 'server.key'
        self.ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        self.ssl_context.load_cert_chain(str(ssl_cert), str(ssl_key))

    @property
    def host_mapping(self):
        """
        provide host <-> port mapping for this server
        """
        return {self.host: self.port}

    async def start(self):
        """
        Start web server
        """
        self.port = unused_port()
        self.handler = self.app.make_handler()
        self.server = await self.loop.create_server(self.handler,
                                                    '127.0.0.1', self.port,
                                                    ssl=self.ssl_context)

    async def stop(self):
        """
        Stop web server
        """
        self.server.close()
        await self.server.wait_closed()
        await self.app.shutdown()
        await self.handler.shutdown()
        await self.app.cleanup()


class FakeLoginSfServer(BaseFakeServer):
    """
    Fake SF login server
    """

    def __init__(self, *, loop):
        super().__init__(loop=loop)
        self.host = 'login.salesforce.com'
        self.app.router.add_routes(
            [web.post('/services/oauth2/token', self.retrieve_token)])

    async def retrieve_token(self, _):
        """
        This route always return the same auth token
        """
        return web.json_response({
            'token_type': 'Bearer',
            'instance_url': 'https://foo.salesforce.com',
            'access_token': 'my-token'})


class FakeSfServer(BaseFakeServer):
    """
    Fake cometd SF server: only support one user !!!
    """

    def __init__(self, *, loop):
        super().__init__(loop=loop)
        self.host = 'foo.salesforce.com'
        self.app.router.add_routes(
            [web.get('/services/data/', self.version_history),
             web.post('/cometd/42.0/', self.cometd)])

        self.client_id = None               # Store the current client id
        self.received_messages = []         # Store all received messages
        self.subscribed_channels = set()    # Store subscribed channels
        self._events_id = defaultdict(int)  # Store event id for each channel
        self._futures = deque()             # Future queue used to push result

    def push_result(self, channel, message):
        """
        Push result to connected client
        """
        # We do not push event to not-subscribed channels
        if channel not in self.subscribed_channels:
            return
        # Update event count
        self._events_id[channel] += 1
        event_id = self._events_id[channel]
        # Create a fake date based on pushed message count
        created_date = datetime.datetime(
            2018, 3, 15, 13, 42, sum(self._events_id.values())).isoformat()
        # Create message
        event = {
            "data": {
                "sobject": message,
                "event": {
                    "replayId": event_id,
                    "createdDate": created_date
                }
            },
            "channel": channel
        }
        self.push_event(event)

    def push_event(self, event):
        """
        Push a raw event to connected client
        """
        # Do not add a future if there is already one waiting for result
        if len(self._futures) != 1 or self._futures[0].done():
            self._futures.append(asyncio.Future())
        self._futures[-1].set_result(event)

    async def wait_for_message(self, timeout=None):
        """
        wait for message pushed to the server
        """
        if not self._futures:
            self._futures.append(asyncio.Future())
        # always wait for the first event
        result = await asyncio.wait_for(self._futures[0], timeout=timeout)
        # Remove it and return the result
        self._futures.popleft()
        return [result]

    async def version_history(self, _):
        """
        Provide a fake api version list
        """
        return web.json_response([{'version': f'{i}.0'}
                                  for i in range(35, 43)])

    async def cometd(self, request):
        """
        Simulate CometD connection strategy
        """
        # We should be authenticate
        if request.headers.get('Authorization', None) != 'Bearer my-token':
            return web.Response(status=403)

        # Extract message
        message = await request.json()
        self.received_messages.append(message)
        channel = message['channel']

        # Only handshake can be check at this step
        if channel == '/meta/handshake':
            return await self.handshake()

        # We can only process other message if the handshake has registred
        # a client id
        if message.get('clientId', None) != self.client_id:
            return web.Response(status=403)

        if channel == '/meta/subscribe':
            return await self.subscribe(message)
        if channel == '/meta/unsubscribe':
            return await self.unsubscribe(message)
        elif channel == '/meta/connect':
            return await self.connect(message)
        elif channel == '/meta/disconnect':
            self.client_id = None
            return web.json_response([{'successful': True}])

        # Unknown channel
        return web.Response(status=404)

    async def handshake(self):
        """
        return handshake response
        """
        # Add an event for timeout advice
        self.push_event({'channel': '/meta/connect', 'advice': {'timeout': 3000}})
        self.client_id = '4'
        return web.json_response([{'clientId': self.client_id}])

    async def subscribe(self, message):
        """
        Subscribe to a channel
        """
        self.subscribed_channels.add(message['subscription'])
        return web.json_response([{'successful': True}])

    async def unsubscribe(self, message):
        """
        Unsubscribe from a channel
        """
        self.subscribed_channels.remove(message['subscription'])
        return web.json_response([{'successful': True}])

    async def connect(self, message):
        """
        Connect and wait for response
        """
        try:
            response = await self.wait_for_message(timeout=1)
            return web.json_response(response)
        except asyncio.TimeoutError:
            return web.Response(status=408)



"""
aio_sf_streaming
~~~~~~~~~~~~~~~~

aio_sf_streaming is a simple asyncio Salesforce Streaming API client for
Python 3.6+
"""
import asyncio
import aiohttp
from .core import BaseSalesforceStreaming
from .connectors import (
    BaseConnector,
    PasswordSalesforceStreaming,
    RefreshTokenSalesforceStreaming,
)
from .mixins import (
    TimeoutAdviceMixin,
    AutoVersionMixin,
    ReplayMixin,
    ReplayType,
    AutoReconnectMixin,
    ReSubscribeMixin,
    AllMixin,
)
from .__version__ import __version__


class SimpleSalesforceStreaming(AllMixin, PasswordSalesforceStreaming):  # Password flow
    """
    A simple helper class providing all-in-one functionalities.

    :param username: User login name
    :param password: User password
    :param client_id: OAuth2 client Id
    :param client_secret: Oauth2 client secret
    :param sandbox: If ``True``, the connexion will be made on a sandbox,
        from ``https://test.salesforce.com`` instead of the main login
        route at ``https://login.salesforce.com``.
    :param version: The API version to use. For example ``'42.0'``.
    :param loop: Asyncio loop used
    :param connector: ``aiohttp`` connector used for main session. Mainly used
        for test purpose.
    :param login_connector: ``aiohttp`` connector used during connection. Mainly
        used for test purpose.
    :param retry_sub_duration: Duration between subscribe retry if server is
        too buzy.
    :param retry_factor: Factor amplification between each successive retry
    :param retry_max_duration: Maximum value of the retry duration
    :param retry_max_count: Maximum count of retry, after this count is reach,
        response or exception are propagated.

    **Usage example**::

        class MyClient(SimpleSalesforceStreaming):
            def __init__(self):
                self.replays = []
                super().__init__(username='my-username',
                                 password='my-password',
                                 client_id='my-client-id',
                                 client_secret='my-client-secret')

            async def store_replay_id(self, channel, replay_id, creation_time):
                # We only store replay id without any use
                self.replays.append((channel, replay_id, creation_time))

            async def get_last_replay_id(self, channel):
                # We ask for only use new events
                return EventType.NEW_EVENTS

        async def print_events():
            async with MyClient() as client:
                await client.subscribe('/topic/Foo')
                async for message in client.events():
                    channel = message['channel']
                    print(f"Message received on {channel} : {message}")

        loop = asyncio.get_event_loop()
        loop.run_until_complete(print_event())

    """

    def __init__(
        self,
        username: str,
        password: str,
        client_id: str,
        client_secret: str,
        *,
        sandbox: bool = False,
        version: str = "42.0",
        loop: asyncio.AbstractEventLoop = None,
        connector: aiohttp.BaseConnector = None,
        login_connector: aiohttp.BaseConnector = None,
        retry_sub_duration: float = 0.1,
        retry_factor: float = 1.,
        retry_max_duration: float = 30.,
        retry_max_count: int = 20,
    ) -> None:
        super().__init__(
            username=username,
            password=password,
            client_id=client_id,
            client_secret=client_secret,
            sandbox=sandbox,
            version=version,
            loop=loop,
            connector=connector,
            login_connector=login_connector,
            retry_sub_duration=retry_sub_duration,
            retry_factor=retry_factor,
            retry_max_duration=retry_max_duration,
            retry_max_count=retry_max_count,
        )


class SimpleRefreshTokenSalesforceStreaming(
    AllMixin, RefreshTokenSalesforceStreaming
):  # Refresh token flow
    """
    A simple helper class providing all-in-one functionalities.

    :param refresh_token: Refresh token
    :param client_id: OAuth2 client Id
    :param client_secret: Oauth2 client secret
    :param sandbox: If ``True``, the connexion will be made on a sandbox,
        from ``https://test.salesforce.com`` instead of the main login
        route at ``https://login.salesforce.com``.
    :param version: The API version to use. For example ``'42.0'``.
    :param loop: Asyncio loop used
    :param connector: ``aiohttp`` connector used for main session. Mainly used
        for test purpose.
    :param login_connector: ``aiohttp`` connector used during connection. Mainly
        used for test purpose.
    :param retry_sub_duration: Duration between subscribe retry if server is
        too buzy.
    :param retry_factor: Factor amplification between each successive retry
    :param retry_max_duration: Maximum value of the retry duration
    :param retry_max_count: Maximum count of retry, after this count is reach,
        response or exception are propagated.

    **Usage example**::

        class MyClient(SimpleRefreshTokenSalesforceStreaming):
            def __init__(self):
                self.replays = []
                super().__init__(refresh_token='refresh_token',
                                 client_id='my-client-id',
                                 client_secret='my-client-secret')

            async def store_replay_id(self, channel, replay_id, creation_time):
                # We only store replay id without any use
                self.replays.append((channel, replay_id, creation_time))

            async def get_last_replay_id(self, channel):
                # We ask for only use new events
                return EventType.NEW_EVENTS

        async def print_events():
            async with MyClient() as client:
                await client.subscribe('/topic/Foo')
                async for message in client.events():
                    channel = message['channel']
                    print(f"Message received on {channel} : {message}")

        loop = asyncio.get_event_loop()
        loop.run_until_complete(print_event())

    """

    def __init__(
        self,
        refresh_token: str,
        client_id: str,
        client_secret: str,
        *,
        sandbox: bool = False,
        version: str = "42.0",
        loop: asyncio.AbstractEventLoop = None,
        connector: aiohttp.BaseConnector = None,
        login_connector: aiohttp.BaseConnector = None,
        retry_sub_duration: float = 0.1,
        retry_factor: float = 1.,
        retry_max_duration: float = 30.,
        retry_max_count: int = 20,
    ) -> None:
        super().__init__(
            refresh_token=refresh_token,
            client_id=client_id,
            client_secret=client_secret,
            sandbox=sandbox,
            version=version,
            loop=loop,
            connector=connector,
            login_connector=login_connector,
            retry_sub_duration=retry_sub_duration,
            retry_factor=retry_factor,
            retry_max_duration=retry_max_duration,
            retry_max_count=retry_max_count,
        )



"""
Misc utils that reproduce iterating tool for coroutine
"""


async def async_enumerate(aiteror, start=0):
    """
    Simple enumerate
    """
    i = start
    async for value in aiteror:
        yield i, value
        i += 1



# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'aio_sf_streaming'
DESCRIPTION = ("aio_sf_streaming is a simple Python 3.6 asyncio library "
               "allowing to connect and receive live notifications from "
               "Salesforce.")
KEYWORDS = "asyncio aiohttp streaming saleforce push notification"
URL = 'https://github.com/papernest-public/aio_sf_streaming'
EMAIL = 'christophe.gabard@papernest.com'
AUTHOR = 'Christophe Gabard / papernest'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = None

REQUIRED = [
    'aiohttp>=3,<4',
]
TESTS_REQUIRED = [
    'pytest>=3,<4',
    'pytest-cov',
    'pytest-asyncio',
    'asynctest',
    'coverage']
DOCS_REQUIRED = [
    'sphinx',
    'sphinx-autobuild',
    'sphinxcontrib-asyncio',
    'sphinx_autodoc_typehints',
]


here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    keywords=KEYWORDS,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    install_requires=REQUIRED,
    extras_require={'tests': TESTS_REQUIRED, 'docs': DOCS_REQUIRED},
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



"""
Unit tests for BaseSalesforceStreaming flow methods
"""
import asyncio
from unittest.mock import call

import aiohttp
from asynctest import patch, CoroutineMock
import pytest

from aio_sf_streaming import BaseSalesforceStreaming
from ..utils.async_itertools import async_enumerate


class SfStreamingTestClass(BaseSalesforceStreaming):
    """
    A fake sf streaming derivated class that always return a fake token
    """
    TEST_ACCESS_TOKEN = "42"
    TEST_INSTANCE_URL = 'https://my-instance.com'

    async def fetch_token(self):
        return self.TEST_ACCESS_TOKEN, self.TEST_INSTANCE_URL


@patch('aio_sf_streaming.BaseSalesforceStreaming.create_connected_session')
@patch('aio_sf_streaming.BaseSalesforceStreaming.handshake')
@pytest.mark.asyncio
async def test_start(mock_handshake, mock_create_session):
    """
    Test start method : should create a session and perform handshake
    """

    client = SfStreamingTestClass()
    await client.start()

    assert mock_create_session.call_count == 1
    assert mock_create_session.call_args == call()

    assert mock_handshake.call_count == 1
    assert mock_handshake.call_args == call()


@patch('aio_sf_streaming.BaseSalesforceStreaming.get_subscribe_payload')
@patch('aio_sf_streaming.BaseSalesforceStreaming.send')
@pytest.mark.asyncio
async def test_subscribe(mock_send, mock_sub_payload):
    """
    Test subscribe method : should retrieve payload, send subscribtion message
    and return the response
    """
    sub_payload = {"channel": "foo"}
    mock_sub_payload.return_value = sub_payload

    send_response = {"succeful": True}
    mock_send.return_value = send_response

    client = SfStreamingTestClass()
    ret = await client.subscribe("foo")

    assert mock_sub_payload.call_count == 1
    assert mock_sub_payload.call_args == call("foo")

    assert mock_send.call_count == 1
    assert mock_send.call_args == call(sub_payload)

    assert ret == send_response


@patch('aio_sf_streaming.BaseSalesforceStreaming.send')
@pytest.mark.asyncio
async def test_messages(mock_send):
    """
    Test messages method : should send connect message and yield results.
    Should reconnect after yield result or timeouts.
    """
    messages = [
        {"channel": '/topic/Foo0'},
        {"channel": '/meta/connect'},
        {"channel": '/topic/Foo1'},
        {"channel": '/topic/Foo2'},
        {"channel": '/topic/Foo3'},
        {"channel": '/topic/Foo4'},
        {"channel": '/meta/connect'},
        {"channel": '/topic/Foo5'},
        {"channel": '/meta/connect'},
    ]
    mock_send.side_effect = [
        # First message : a simple success
        [messages[0]],
        # Timeout
        aiohttp.ClientResponseError(None, None, code=408),
        # Second message : multiples messages
        messages[1:4],
        # Timeout
        asyncio.TimeoutError(),
        # Other messages
        messages[4:]]

    client = SfStreamingTestClass()
    received_messages = []
    # We iterate message and ask stop after received 7 messages
    async for i, m in async_enumerate(client.messages()):
        received_messages.append(m)
        if i == 6:
            await client.ask_stop()

    # We should have only 7 messages
    assert received_messages == messages[:7]
    assert mock_send.call_count == 5

    client = SfStreamingTestClass()
    # An other error shoud be re-raised
    mock_send.side_effect = [
        aiohttp.ClientResponseError(None, None, code=404),
    ]
    with pytest.raises(aiohttp.ClientResponseError):
        async for m in client.messages():
            # Avoid infinite loop if test fail
            assert False


@patch('aio_sf_streaming.BaseSalesforceStreaming.send')
@pytest.mark.asyncio
async def test_events(mock_send):
    """
    Test events method : same as messages but skip 'meta' messages
    """
    messages = [
        {"channel": '/topic/Foo0'},
        {"channel": '/meta/connect'},
        {"channel": '/topic/Foo1'},
        {"channel": '/topic/Foo2'},
        {"channel": '/topic/Foo3'},
        {"channel": '/topic/Foo4'},
        {"channel": '/meta/connect'},
        {"channel": '/topic/Foo5'},
        {"channel": '/meta/connect'},
    ]
    mock_send.side_effect = [
        # First message : a simple success
        [messages[0]],
        # Second message : multiples messages
        messages[1:4],
        # Timeout
        asyncio.TimeoutError(),
        # Other messages
        messages[4:]]

    client = SfStreamingTestClass()
    received_messages = []
    # We iterate message and ask stop after received 6 messages
    async for i, m in async_enumerate(client.events()):
        received_messages.append(m)
        if i == 5:
            await client.ask_stop()

    # We should have only 6 messages, thus without meta
    excepted_messages = [m for m in messages if m["channel"] != '/meta/connect']
    assert received_messages == excepted_messages
    assert mock_send.call_count == 4


@patch('aio_sf_streaming.BaseSalesforceStreaming.get_unsubscribe_payload')
@patch('aio_sf_streaming.BaseSalesforceStreaming.send')
@pytest.mark.asyncio
async def test_unsubscribe(mock_send, mock_unsub_payload):
    """
    Test unsubscribe method : should retrieve payload, send unsubscribtion message
    and return the response
    """
    unsub_payload = {"channel": "foo"}
    mock_unsub_payload.return_value = unsub_payload

    send_response = {"succeful": True}
    mock_send.return_value = send_response

    client = SfStreamingTestClass()
    ret = await client.unsubscribe("foo")

    assert mock_unsub_payload.call_count == 1
    assert mock_unsub_payload.call_args == call("foo")

    assert mock_send.call_count == 1
    assert mock_send.call_args == call(unsub_payload)

    assert ret == send_response


@patch('aio_sf_streaming.BaseSalesforceStreaming.ask_stop')
@patch('aio_sf_streaming.BaseSalesforceStreaming.disconnect')
@patch('aio_sf_streaming.BaseSalesforceStreaming.close_session')
@pytest.mark.asyncio
async def test_close(mock_close_session, mock_disconnect, mock_ask_stop):
    """
    Test stop method : should ask stopping, disconnect from cometd and close
    connection
    """
    client = SfStreamingTestClass()
    await client.stop()

    assert mock_ask_stop.call_count == 1
    assert mock_ask_stop.call_args == call()

    assert mock_disconnect.call_count == 1
    assert mock_disconnect.call_args == call()

    assert mock_close_session.call_count == 1
    assert mock_close_session.call_args == call()


def test_token_url():
    """
    Test token url property
    """
    assert (SfStreamingTestClass(sandbox=True).token_url ==
            'https://test.salesforce.com/services/oauth2/token')
    assert (SfStreamingTestClass(sandbox=False).token_url ==
            'https://login.salesforce.com/services/oauth2/token')
    assert (SfStreamingTestClass().token_url ==
            'https://login.salesforce.com/services/oauth2/token')


@patch('aiohttp.ClientSession')
@pytest.mark.asyncio
async def test_create_connected_session(mock_client_session):
    """
    Test create_connected_session: Should create an aiohttp client session
    using provided session and set provided instance_url
    """
    client = SfStreamingTestClass()
    result = await client.create_connected_session()

    assert client.instance_url == SfStreamingTestClass.TEST_INSTANCE_URL

    assert mock_client_session.call_count == 1
    _, kwargs = mock_client_session.call_args
    assert kwargs['headers']['Authorization'] == 'Bearer 42'

    assert result is mock_client_session()


@pytest.mark.asyncio
async def test_close_session():
    """
    Test close session: Call close session if defined
    """
    client = SfStreamingTestClass()
    with patch.object(client, 'session') as mock_session:
        mock_session_close = CoroutineMock()
        mock_session.close = mock_session_close
        await client.close_session()

    assert mock_session_close.call_count == 1
    assert client.session is None

    # An other one to check that it does not generate errors
    await client.close_session()
    assert mock_session_close.call_count == 1
    assert client.session is None


def test_end_point():
    """
    Test end point property depending of the provided version
    """
    assert SfStreamingTestClass(version='1.0').end_point == '/cometd/1.0/'
    assert SfStreamingTestClass(version='42.0').end_point == '/cometd/42.0/'
    assert SfStreamingTestClass().end_point == '/cometd/42.0/'


@pytest.mark.asyncio
async def test_handshake_payload():
    """
    Test handshake_payload
    """
    client = SfStreamingTestClass()
    # the handshake payload is constant
    payload = await client.get_handshake_payload()

    assert payload == {
            'channel': '/meta/handshake',
            'supportedConnectionTypes': ['long-polling'],
            'version': '1.0',
            'minimumVersion': '1.0'}


@pytest.mark.asyncio
async def test_subscribe_payload():
    """
    Test subscribe payload
    """
    client = SfStreamingTestClass()
    payload = await client.get_subscribe_payload('/topic/Foo')

    assert payload == {
        'channel': '/meta/subscribe',
        'subscription': '/topic/Foo'
    }


@pytest.mark.asyncio
async def test_unsubscribe_payload():
    """
    Test unsubscribe payload
    """
    client = SfStreamingTestClass()
    payload = await client.get_unsubscribe_payload('/topic/Foo')

    assert payload == {
        'channel': '/meta/unsubscribe',
        'subscription': '/topic/Foo'
    }


@patch('aio_sf_streaming.BaseSalesforceStreaming.send')
@pytest.mark.asyncio
async def test_handhshake(mock_send):
    """
    Test handshake: should send handshake_payload, set client id and return
    provided response
    """
    send_response = [{"succeful": True, 'clientId': 'Foo'}]
    mock_send.return_value = send_response

    client = SfStreamingTestClass()
    ret = await client.handshake()

    assert mock_send.call_count == 1
    assert mock_send.call_args == call({
            'channel': '/meta/handshake',
            'supportedConnectionTypes': ['long-polling'],
            'version': '1.0',
            'minimumVersion': '1.0'})

    assert ret == send_response
    assert client.client_id == 'Foo'


@patch('aio_sf_streaming.BaseSalesforceStreaming.send')
@pytest.mark.asyncio
async def test_disconnect(mock_send):
    """
    Test disconnect: should send diconnect
    """
    send_response = {"succeful": True}
    mock_send.return_value = send_response

    client = SfStreamingTestClass()
    ret = await client.disconnect()

    assert mock_send.call_count == 1
    assert mock_send.call_args == call({'channel': '/meta/disconnect'})

    assert ret == send_response


@patch('aio_sf_streaming.BaseSalesforceStreaming.post')
@pytest.mark.asyncio
async def test_send(mock_post):
    """
    Test send: complete payload and post response
    """
    post_response = {"succeful": True}
    mock_post.return_value = post_response

    # Without client_id
    client = SfStreamingTestClass()
    client.message_count = 41
    client.client_id = None
    ret = await client.send({'foo': 'bar'})

    assert mock_post.call_count == 1
    assert mock_post.call_args == call(client.end_point,
                                       json={'foo': 'bar', 'id': '42'})
    assert ret == post_response

    # With client_id
    assert client.message_count == 42
    client.client_id = 'buzz'
    ret = await client.send({'foo2': 'bar2'})

    assert mock_post.call_count == 2
    assert mock_post.call_args == call(client.end_point,
                                       json={'foo2': 'bar2',
                                             'clientId': 'buzz',
                                             'id': '43'})
    assert ret == post_response


@patch('aio_sf_streaming.BaseSalesforceStreaming.request')
@pytest.mark.asyncio
async def test_get(mock_request):
    """
    Test get: helper for request
    """
    request_response = {"succeful": True}
    mock_request.return_value = request_response

    client = SfStreamingTestClass()
    response = await client.get('/foo', params={'q': 'test'})

    assert response == request_response
    assert mock_request.call_count == 1
    assert mock_request.call_args == call('get', '/foo', params={'q': 'test'})


@patch('aio_sf_streaming.BaseSalesforceStreaming.request')
@pytest.mark.asyncio
async def test_post(mock_request):
    """
    Test post: helper for request
    """
    request_response = {"succeful": True}
    mock_request.return_value = request_response

    client = SfStreamingTestClass()
    response = await client.post('/foo', json={'q': 'test'})

    assert response == request_response
    assert mock_request.call_count == 1
    assert mock_request.call_args == call('post', '/foo', json={'q': 'test'})


@patch('aio_sf_streaming.BaseSalesforceStreaming.start')
@patch('aio_sf_streaming.BaseSalesforceStreaming.stop')
@pytest.mark.asyncio
async def test_ctx_manager(mock_stop, mock_start):
    """
    Test context manager interface
    """

    async with SfStreamingTestClass() as client:
        assert mock_start.call_count == 1
        assert isinstance(client, SfStreamingTestClass)
        assert mock_stop.call_count == 0
    assert mock_stop.call_count == 1

    try:
        async with SfStreamingTestClass() as client:
            assert mock_start.call_count == 2
            assert mock_stop.call_count == 1
            raise Exception()
    except Exception:
        assert mock_stop.call_count == 2
